##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Ape configuration assembler.

$Id: apeconf.py,v 1.1 2004/06/28 18:08:29 johanc Exp $
"""
from __future__ import nested_scopes
from types import TupleType

from apelib.core.mapper import Mapper, MapperConfiguration
from apelib.core.serializers import CompositeSerializer, AnyObjectSerializer
from apelib.core.gateways import CompositeGateway
from apelib.core.interfaces import IDatabaseInitializer, ConfigurationError

from apelib.zodb3.zodbtables import Table, TableSchema
from common import Directive, DirectiveReader, ComponentSystem


class MapperDeclaration(Directive):
    schema = TableSchema()
    schema.add('mapper_name', primary=1)


class MapperAttribute(Directive):
    schema = TableSchema()
    schema.add('mapper_name', primary=1)
    # Attribute names: 'class', 'extends', 'register-class'
    schema.add('name', primary=1, indexed=1)
    schema.add('value')


class ComponentDefinition(Directive):
    schema = TableSchema()
    # comptypes: 'serializer', 'gateway', 'classifier', 'oid_generator'
    schema.add('comptype', primary=1)
    schema.add('name', primary=1)
    schema.add('producer')


class MapperComponent(Directive):
    schema = TableSchema()
    schema.add('mapper_name', primary=1)
    schema.add('comptype', primary=1)
    schema.add('name', primary=1)
    schema.add('producer')
    schema.add('order')


class MapperUseFor(Directive):
    schema = TableSchema()
    schema.add('mapper_name', primary=1)
    schema.add('attr', primary=1)
    schema.add('value', primary=1)  # Multiple values allowed


class ClassifierOption(Directive):
    schema = TableSchema()
    schema.add('mapper_name', primary=1)
    schema.add('option', primary=1)
    schema.add('value')


class DisabledProducer:
    def __init__(self, source):
        self.source = source

    def __call__(self, compsys):
        return None


class UseProducer:
    def __init__(self, source, comptype, name):
        self.source = source
        self.comptype = comptype
        self.name = name

    def __call__(self, compsys):
        return compsys.get(self.comptype, self.name)


class FactoryProducer:

    def __init__(self, source, factory):
        self.source = source
        pos = factory.find('(')
        if pos >= 0:
            # Arguments specified.  Interpret as a Python expression.
            args = eval(factory[pos:])
            if not isinstance(args, TupleType):
                args = (args,)
            factory = factory[:pos]
        else:
            args = ()
        pos = factory.rfind('.')
        if pos <= 0:
            raise ValueError, "Module and class name required"
        self.module_name = factory[:pos]
        self.class_name = factory[pos + 1:]
        self.args = args
        self.sub_producer = None

    def __call__(self, compsys):
        params = []
        if self.args:
            params.extend(self.args)
        if self.sub_producer is not None:
            obj = self.sub_producer(compsys)
            params.append(obj)
        m = __import__(self.module_name, {}, {}, ('__doc__',))
        try:
            c = getattr(m, self.class_name)
        except AttributeError:
            raise ImportError, "No class %s in module %s" % (
                self.class_name, self.module_name)
        return c(*params)


def make_producer(source, comptype, attrs, raise_exc=1):
    if attrs.get('enabled', '').lower() == 'false':
        return DisabledProducer(source)
    elif attrs.has_key('use'):
        if attrs.has_key('factory'):
            raise ValueError, "Both 'use' and 'factory' not allowed"
        return UseProducer(source, comptype, attrs['use'])
    elif attrs.has_key('factory'):
        return FactoryProducer(source, attrs['factory'])
    elif raise_exc:
        raise ValueError, "Need a 'use', 'factory', or 'enabled' attribute"
    else:
        return None


def get_element_handlers():
    """Returns a dictionary of XML element handlers.
    """

    def handle_configuration(source, vars, attrs):
        assert vars.has_key('directives')

    def handle_variation(source, vars, attrs):
        variations = vars['variations']
        name = attrs['name']
        d = variations.get(name)
        if d is None:
            # Make a new variation.
            d = []
            variations[name] = d
        vars['directives'] = d

    def handle_mapper(source, vars, attrs):
        d = vars['directives']
        mapper_name = str(attrs['name'])
        vars['mapper_name'] = mapper_name
        d.append(MapperDeclaration(source, mapper_name))
        for key in ('class', 'extends', 'register-class'):
            if attrs.has_key(key):
                value = attrs[key]
                if (key == 'register-class'):
                    if value.lower() == 'true':
                        value = True
                    elif value.lower() == 'false':
                        value = False
                    else:
                        raise ValueError("Value must be true or false")
                d.append(MapperAttribute(source, mapper_name, key, value))

    def handle_mapper_component(source, vars, attrs, comptype):
        d = vars['directives']
        producer = make_producer(source, comptype, attrs)
        mapper_name = vars.get('mapper_name')
        if mapper_name is None:
            # Reusable component
            name = attrs['name']
            directive = ComponentDefinition(
                source, comptype, name, producer)
        else:
            # Composite component of a mapper
            name = attrs.get('name', '')
            directive = MapperComponent(
                source, mapper_name, comptype,
                name, producer, attrs.get('order', 'middle'))
        d.append(directive)
        return producer

    def handle_serializer(source, vars, attrs):
        handle_mapper_component(source, vars, attrs, 'serializer')

    def handle_gateway(source, vars, attrs):
        p = vars.get('classifier_producer')
        if p is not None:
            # Set a gateway for a classifier.
            if not hasattr(p, 'sub_producer'):
                raise ValueError(
                    "Classifier at %s needs a factory in order to "
                    "use a gateway" % source)
            if p.sub_producer is not None:
                raise ValueError(
                    "Multiple gateways in classifiers not allowed at %s" %
                    source)
            p.sub_producer = make_producer(source, 'gateway', attrs)
        else:
            handle_mapper_component(source, vars, attrs, 'gateway')

    def handle_classifier(source, vars, attrs):
        d = vars['directives']
        producer = make_producer(source, 'classifier', attrs)
        directive = ComponentDefinition(source, 'classifier', '', producer)
        d.append(directive)
        vars['classifier_producer'] = producer

    def handle_oid_generator(source, vars, attrs):
        d = vars['directives']
        producer = make_producer(source, 'oid_generator', attrs)
        directive = ComponentDefinition(source, 'oid_generator', '', producer)
        d.append(directive)

    def handle_use_for(source, vars, attrs):
        d = vars['directives']
        mapper_name = vars['mapper_name']
        for attr in ('class', 'extensions', 'generic', 'oid'):
            if attrs.has_key(attr):
                v = attrs[attr]
                if attr == 'extensions':
                    first = 1
                    for ext in v.split():
                        if not ext.startswith('.'):
                            ext = '.' + ext
                        ext = ext.lower()
                        d.append(MapperUseFor(
                            source, mapper_name, 'extension', ext))
                        if first:
                            # Use a classifier option to set the default
                            # extension.
                            first = 0
                            d.append(ClassifierOption(
                                source, mapper_name, 'default_extension', ext))
                else:
                    d.append(MapperUseFor(
                        source, mapper_name, attr, v))

    def handle_option(source, vars, attrs):
        d = vars['directives']
        mapper_name = vars['mapper_name']
        name = attrs['name']
        value = attrs['value']
        d.append(ClassifierOption(source, mapper_name, name, value))

    handlers = {
        'configuration': handle_configuration,
        'variation':     handle_variation,
        'mapper':        handle_mapper,
        'serializer':    handle_serializer,
        'gateway':       handle_gateway,
        'classifier':    handle_classifier,
        'oid-generator': handle_oid_generator,
        'use-for':       handle_use_for,
        'option':        handle_option,
        }

    return handlers



class BasicComponentAssembler:
    """Assembler for producer-based components.

    Configures at the time of creation.
    """

    def __init__(self, compsys, comptype, name):
        self.compsys = compsys
        records = compsys.dtables.query(
            ComponentDefinition, comptype=comptype, name=name)
        if not records:
            raise ConfigurationError("No %s component named %s exists"
                                     % (comptype, repr(name)))
        assert len(records) == 1
        producer = records[0]['producer']
        self.producer = producer

    def create(self):
        return self.producer(self.compsys)

    def configure(self):
        pass


class MapperAssembler:
    """Assembler for one mapper component.
    """
    def __init__(self, compsys, comptype, name):
        self.compsys = compsys
        dtables = compsys.dtables
        self.mapper_name = name
        if not dtables.query(MapperDeclaration, mapper_name=name):
            raise ConfigurationError("No mapper named %s exists" % repr(name))
        self.subobjs = []  # all subobjects
        self.attrs = {}
        for record in dtables.query(MapperAttribute, mapper_name=name):
            self.attrs[record.name] = record.value
        self._prepare_sub_components()

    def _prepare_sub_components(self):
        self.multi_comps = {}  # comptype -> name -> record
        dtables = self.compsys.dtables
        name = self.mapper_name
        all_names = []  # mapper_name and all of its base mapper_names
        while name:
            all_names.append(name)
            records = dtables.query(
                MapperComponent, mapper_name=name)
            for r in records:
                d = self.multi_comps.setdefault(r.comptype, {})
                d.setdefault(r.name, r)
            name = dtables.query_field(
                MapperAttribute, 'value', mapper_name=name, name='extends')
            if name and name in all_names:
                raise ConfigurationError(
                    "Circular extension in mappers %s" % repr(all_names))

    def create(self):
        self.obj = Mapper()
        return self.obj

    def configure(self):
        self.add_serializers()
        self.add_gateways()
        self.add_initializers()

    def add_serializers(self):
        cname = self.attrs.get('class')
        if cname == 'none':
            # This mapper is abstract (usable for no classes)
            module, name = None, None
        else:
            # This mapper is concrete (usable for one class only)
            if cname is None:
                cname = self.mapper_name
            pos = cname.rfind('.')
            if pos < 0:
                raise ConfigurationError("Class name must include a module name")
            module, name = cname[:pos], cname[pos + 1:]
        d = self.multi_comps.get('serializer', {})

        # Create the main serializer
        r = d.get('')
        if r:
            s = r.producer(self.compsys)
        else:
            s = CompositeSerializer(module, name)

        # Create the contained serializers
        ordered = [
            ((r.order or '').lower(), name, r)
            for name, r in d.items() if name]
        ordered.sort()
        for order, name, r in ordered:
            o = r.producer(self.compsys)
            if o is not None:
                s.add(str(name), o)
                self.subobjs.append(o)

        # Assign it
        self.obj.serializer = s

    def add_gateways(self):
        d = self.multi_comps.get('gateway', {})

        # Create the main gateway
        r = d.get('')
        if r:
            g = r.producer(self.compsys)
        else:
            g = CompositeGateway()

        # Create the contained gateways
        for name, r in d.items():
            if name:
                o = r.producer(self.compsys)
                if o is not None:
                    g.add(str(name), o)
                    self.subobjs.append(o)

        # Assign it
        self.obj.gateway = g

    def add_initializers(self):
        for o in self.subobjs:
            if IDatabaseInitializer.isImplementedBy(o):
                self.obj.initializers.append(o)


class ClassifierAssembler (BasicComponentAssembler):
    """Assembler for one classifier.
    """
    def __init__(self, compsys, comptype, name):
        assert comptype == "classifier", comptype
        assert name == '', name
        BasicComponentAssembler.__init__(self, compsys, comptype, name)

    def create(self):
        self.obj = BasicComponentAssembler.create(self)
        return self.obj

    def configure(self):
        dtables = self.compsys.dtables
        all_regs = {}     # { (attr, value) -> mapper_name }
        all_options = {}  # { (mapper_name, option) -> value }
        mapper_names = [r.mapper_name for r in
                        dtables.query(MapperDeclaration)]
        # Gather classification options from each mapper configuration.
        for name in mapper_names:
            # use-for directives
            records = dtables.query(
                MapperUseFor, mapper_name=name)
            for r in records:
                key = ((r.attr, r.value))
                if all_regs.has_key(key) and all_regs[key] != name:
                    raise ConfigurationError(
                        "Mappers %s and %s are contending over %s == %s" % (
                        name, all_regs[key],
                        repr(r.attr), repr(r.value)))
                all_regs[key] = name

            register_class = dtables.query_field(
                MapperAttribute, 'value', mapper_name=name,
                name='register-class')
            if register_class or register_class is None:
                # Create an implicit 'use-for class' directive
                class_name = dtables.query_field(
                    MapperAttribute, 'value', mapper_name=name, name='class')
                if class_name is None:
                    class_name = name
                elif class_name == 'none':
                    class_name = None
                if class_name is not None:
                    key = ('class', class_name)
                    if all_regs.has_key(key) and all_regs[key] != name:
                        raise ConfigurationError(
                            "Mappers %s and %s are contending over %s == %s" % (
                            name, all_regs[key],
                            "'class'", repr(class_name)))
                    all_regs[key] = name

            # options
            records = dtables.query(
                ClassifierOption, mapper_name=name)
            for r in records:
                all_options[(name, r.option)] = r.value

        # Perform the registrations.
        if all_regs or all_options:
            for (attr, value), name in all_regs.items():
                self.obj.register(attr, value, name)
            for (name, option), value in all_options.items():
                self.obj.set_option(name, option, value)
            
        

def configure(filenames, vname=''):
    """Returns a MapperConfiguration built from configuration files.
    """
    handlers = get_element_handlers()
    reader = DirectiveReader(handlers)
    for fn in filenames:
        reader.read(fn)
    directives = reader.get_directives(vname)
    cs = ComponentSystem(directives)
    cs.add_component_type('mapper', MapperAssembler)
    cs.add_component_type('classifier', ClassifierAssembler)
    for comptype in ('serializer', 'gateway', 'oid_generator'):
        cs.add_component_type(comptype, BasicComponentAssembler)
    mappers = {}
    for record in cs.dtables.query(MapperDeclaration):
        name = record.mapper_name
        mappers[name] = cs.get('mapper', name)
    classifier = cs.get('classifier', '')
    oid_gen = cs.get('oid_generator', '')
    conf = MapperConfiguration(mappers, classifier, oid_gen)
    for obj in (classifier.gateway, oid_gen):
        if IDatabaseInitializer.isImplementedBy(obj):
            conf.initializers.append(obj)
    conf.check()
    return conf

