##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""SQL gateways for security information.

$Id: security.py,v 1.1 2004/06/28 18:08:30 johanc Exp $
"""

from apelib.core.schemas import RowSequenceSchema
from sqlbase import SQLGatewayBase


class SQLSecurityAttributes (SQLGatewayBase):
    """SQL security attribute storage"""

    __implements__ = SQLGatewayBase.__implements__

    schema = RowSequenceSchema()
    schema.add('declaration_type', 'string')
    schema.add('role', 'string')
    schema.add('permission', 'string')
    schema.add('username', 'string')

    table = 'security'
    oid_column_def = (('oid', 'int', 0),)  # Don't create a primary key

    def load(self, event):
        conn = self.get_connection(event)
        items = conn.select(self.table, self.columns, oid=event.oid)
        items.sort()
        return items, tuple(items)

    def store(self, event, state):
        conn = self.get_connection(event)
        conn.set_many(self.table, event.oid, (), self.columns, state)
        state = list(state)
        state.sort()
        return tuple(state)



class SQLUserList (SQLGatewayBase):
    """Stores and retrieves all users for a folder at once."""

    __implements__ = SQLGatewayBase.__implements__

    schema = RowSequenceSchema()
    schema.add('id', 'string', 1)
    schema.add('password', 'string')
    schema.add('roles', 'string:list')
    schema.add('domains', 'string:list')

    table_defs = {
        'users':        (('oid', 'int', 1),
                         ('id', 'string', 1),
                         ('password', 'string', 0),),
        'user_roles':   (('oid', 'int', 0),
                         ('id', 'string', 0),
                         ('role', 'string', 0),),
        'user_domains': (('oid', 'int', 0),
                         ('id', 'string', 0),
                         ('domain', 'string', 0),),
        }


    def init(self, event):
        conn = self.get_connection(event)
        for table, col_defs in self.table_defs.items():
            if not conn.exists(table, 'table'):
                conn.create_table(table, col_defs)
            elif event.clear_all:
                conn.delete_from(table)


    def load(self, event):
        conn = self.get_connection(event)
        rows = conn.select('users', ('id', 'password'), oid=event.oid)
        data = {}
        for id, password in rows:
            data[id] = (password, [], [])
        rows = conn.select('user_roles', ('id', 'role'), oid=event.oid)
        for id, role in rows:
            row = data.get(id)
            if row is not None:
                row[1].append(role)
        rows = conn.select('user_domains', ('id', 'domain'), oid=event.oid)
        for id, domain in rows:
            row = data.get(id)
            if row is not None:
                row[2].append(domain)
        records = []
        for id, (password, roles, domains) in data.items():
            records.append((id, password, tuple(roles), tuple(domains)))
        records.sort()
        return records, tuple(records)


    def store(self, event, state):
        oid = event.oid
        conn = self.get_connection(event)
        rows = [(id, pw) for id, pw, roles, domains in state]
        conn.set_many('users', event.oid, (), ('id', 'password',), rows)
        roles_d = {}
        domains_d = {}
        for id, pw, roles, domains in state:
            for role in roles:
                roles_d[(id, role)] = 1
            for domain in domains:
                domains_d[(id, domain)] = 1
        conn.set_many(
            'user_roles', event.oid, (), ('id', 'role',), roles_d.keys())
        conn.set_many(
            'user_domains', event.oid, (), ('id', 'domain',), domains_d.keys())
        state = list(state)
        state.sort()
        return tuple(state)
