##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""SQL gateways for a filesystem-like tree structure.

$Id: structure.py,v 1.1 2004/06/28 18:08:30 johanc Exp $
"""

from apelib.core.schemas import FieldSchema, RowSequenceSchema
from sqlbase import SQLGatewayBase


class SQLObjectData (SQLGatewayBase):
    """SQL object data gateway"""

    __implements__ = SQLGatewayBase.__implements__

    schema = FieldSchema('data', 'string')
    table = 'object_data'
    column_defs = (
        ('data', 'blob', 0),
        )

    def load(self, event):
        conn = self.get_connection(event)
        firstcol = self.columns[:1]
        items = conn.select(self.table, firstcol, oid=event.oid)
        if items:
            state = items[0][0]
        else:
            state = ''
        return state, state

    def store(self, event, state):
        conn = self.get_connection(event)
        firstcol = (self.column_defs[0][0],)
        data = (conn.module.Binary(state),)
        conn.set_one(self.table, event.oid, firstcol, data, event.is_new)
        return state


class SQLFolderItems (SQLGatewayBase):
    """SQL folder items gateway"""

    __implements__ = SQLGatewayBase.__implements__

    schema = RowSequenceSchema()
    schema.add('key', 'string', 1)
    schema.add('oid', 'string')
    schema.add('classification', 'classification')

    table = 'folder_items'
    column_defs = (
        ('name', 'string', 1),
        ('child_oid', 'int', 0),
        )

    def load(self, event):
        conn = self.get_connection(event)
        rows = conn.select(self.table, self.columns, oid=event.oid)
        res = []
        h = []
        for name, child_oid in rows:
            s = str(child_oid)
            classification = event.classify(s)
            res.append((name, s, classification))
            h.append((name, long(child_oid)))
        h.sort()
        return res, tuple(h)

    def store(self, event, state):
        conn = self.get_connection(event)
        rows = [(name, long(child_oid)) for (name, child_oid, cls) in state]
        rows.sort()
        # Note that set_many() requires the child_oid column to match
        # its database type.
        conn.set_many(self.table, event.oid, ('name',), ('child_oid',), rows)
        return tuple(rows)


class SQLItemId (SQLGatewayBase):
    """SQL item ID gateway.

    Piggybacks SQLFolderItems for init and store.
    Makes the assumption that the item is stored in only one place.
    """

    __implements__ = SQLGatewayBase.__implements__

    schema = FieldSchema('id', 'string')
    table = 'folder_items'

    column_defs = (
        ('child_oid', 'int', 1),
        ('name', 'string', 0),
        )

    def init(self, event):
        pass

    def load(self, event):
        conn = self.get_connection(event)
        rows = conn.select(self.table, ('name',), child_oid=event.oid)
        if len(rows) >= 1:
            name = rows[0][0]  # Accept only the first result
        else:
            name = None
        # Disable conflict checking by returning None as the hash value.
        return name, None

    def store(self, event, state):
        return None


class SQLRemainder (SQLObjectData):
    """SQL remainder pickle gateway"""

    __implements__ = SQLGatewayBase.__implements__

    table = 'remainder'
    column_defs = (
        ('pickle', 'blob', 0),
        )


class SQLModTime (SQLGatewayBase):
    """SQL object mod time gateway"""

    __implements__ = SQLGatewayBase.__implements__

    schema = FieldSchema('mtime', 'int')  # second
    table = 'mtime'
    column_defs = (
        ('mtime', 'long', 0),
        )

    def load(self, event):
        conn = self.get_connection(event)
        items = conn.select(self.table, self.columns, oid=event.oid)
        if items:
            state = long(items[0][0])
        else:
            state = 0L
        return state, state

    def store(self, event, state):
        state = long(state)
        conn = self.get_connection(event)
        data = (state,)
        conn.set_one(self.table, event.oid, self.columns, data, event.is_new)
        return state


def root_mapping():
    """Returns a gateway suitable for storing the root persistent mapping.
    """
    from apelib.core.gateways import CompositeGateway
    g = CompositeGateway()
    g.add('references', SQLFolderItems())
    g.add('others', SQLObjectData())
    return g
