##############################################################################
#
# Copyright (c) 2004 Chris McDonough, Paul Winkler, Jens Vagelpohl and
# Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os
import sys
import time
import tempfile
import md5

import zLOG

from thread import allocate_lock

from AccessControl.SecurityManagement import getSecurityManager
from BTrees import OOBTree
from ZPublisher.Iterators import filestream_iterator
from Products.StandardCacheManagers.RAMCacheManager import RAMCache
from Products.PageTemplates.Expressions import getEngine, \
     SecureModuleImporter
from Products.PageTemplates.TALES import CompilerError

_debug = 0

class FileCache(RAMCache):
    """ A cache that caches rendered content to the filesystem """
    _tempfile_path = ''

    def __init__(self, path='/tmp'):
        # self.cache maps physical ZODB paths to disk files.
        self.cache=OOBTree.OOBTree()
        self.next_cleanup = 0
        self.setDir(path)
        self._makeDirs()
        self._naming_expr = None
        self._tempfile_path = ''
        self.entries = {}

    def _fileName(self, ob):
        """ Compute a filename based on an MD5 hash: doesn't preserve
        human-readable path, but otherwise makes life much easier """
        if self._naming_expr is not None:
            expr_context = self.getExprContext(ob)
            munged = self._naming_expr(expr_context)
            fname = os.path.join(self._dir, munged)
        else:
            phys_path = '/'.join(ob.getPhysicalPath())[1:]
            hashed = md5.new(phys_path).hexdigest()
            fname = os.path.join(self._dir, hashed[:2], hashed[:3], hashed)

        return fname

    def getExprContext(self, ob=None):
        """ Create a context for evaluating the TAL naming expression """
        data =  { 'here'       : self
                , 'user'       : getSecurityManager().getUser()
                , 'request'    : getattr(ob, 'REQUEST', None)
                , 'modules'    : SecureModuleImporter
                , 'object'     : ob
                , 'object_url' : ob.absolute_url()
                , 'nothing'    : None
                }

        return getEngine().getContext(data)

    def setNamingExpression(self, naming_expression=''):
        """ Set a new file naming expression """
        if naming_expression:
            self._naming_expr = getEngine().compile(naming_expression)
        else:
            self._naming_expr = None

    def setTempfileDir(self, path=''):
        self._tempfile_path = path

    def getDir(self):
        """ Retrieve the filesystem directory used for caching """
        return self._dir

    def getTempfileDir(self):
        """ Retrieve the filesystem directory used for caching """
        return self._tempfile_path

    def setDir(self, dir):
        """ Set the filesystem diriectory to use for caching """
        self._dir = dir
        self._makeDirs()

    def setTempfileDir(self, dir):
        """ Retrieve the filesystem directory used for caching """
        self._tempfile_path = dir

    def ZCache_invalidate(self, ob):
        """ Invalidate cache entries that apply to ob """
        if _debug:
            print "ZCache_invalidate called"
        fname = self._fileName(ob)
        cache_entry = FileCacheEntry(self, ob, fname, self._tempfile_path)
        cache_entry.remove()

    def ZCache_get( self, ob, view_name='', keywords=None
                  , mtime_func=None, default=None
                  ):
        """ Gets a cache entry and return a filestream_iterator """
        pp = '/'.join(ob.getPhysicalPath())
        fname = self._fileName(ob)

        try:
            fiter = filestream_iterator(fname, 'rb')
        except IOError:
            # couldn't get the actual cached file
            msg = ('Failed to retrieve cached file for "%s" using "%s" '
                   'as a filename' % (pp, fname))
            zLOG.LOG('FileCacheManager', zLOG.INFO, msg, error=sys.exc_info())
            return default

        if _debug:
            print "returning filestream iterator for %s" % pp

        try:
            entry = self.entries[pp]
        except KeyError:
            entry = {'path':pp, 'fpath':fname, 'size':0, 'hits':0}
        entry['hits'] += 1
        self.entries[pp] = entry

        return fiter

    def ZCache_set( self, ob, data=None, view_name=''
                  , keywords=None, mtime_func=None
                  ):
        """ Sets a cache entry. """
        if _debug:
            print "ZCache_set called"
        fname = self._fileName(ob)

        if data is None:
            # maybe it's a File or an Image, grab the data
            data = ob.data

        cache_entry = FileCacheEntry(self, ob, fname, self._tempfile_path)
        cache_entry.write(data)

    def record(self, state):
        pp = '/'.join(state.ob.getPhysicalPath())
        if state.op_remove:
            try:
                del self.entries[pp]
            except KeyError:
                pass
        elif state.op_write:
            empty = {'path':pp, 'fpath':state.path, 'size':0, 'hits':0}
            entry = self.entries.setdefault(pp, empty)
            entry['path'] = pp
            entry['size'] = state.size
            entry['fpath'] = state.path

    def values(self):
        return self.entries.values()

    def _makeDirs(self):
        """ Make sure we have somewhere to put files. """
        chars = '0123456789abcdef'

        for char1 in chars:
            for char2 in chars:
                for char3 in chars:
                    dirpath = os.path.join(self._dir, '%s%s' % (char1, char2),
                                           '%s%s%s' % (char1, char2, char3))
                    try:
                        os.makedirs(dirpath)
                    except OSError, why:
                        # it might have already been created
                        if not os.path.isdir(dirpath):
                            zLOG.LOG('FileCacheManager', zLOG.PROBLEM,
                                     'unable to create directory %s' % dirpath,
                                     error=sys.exc_info())


class FileCacheEntry:
    """ Class representing a cache file """

    makedirs = True # make directories as necessary

    def __init__(self, cache, ob, path, tmpdir):
        """ Instantiate a new instance """
        self.cache = cache
        self.ob = ob
        self.path = path
        self.tmpdir = tmpdir
        self.temppath = None
        self.writelock = allocate_lock()
        self.op_remove = 0
        self.op_write = 0
        self._transaction_done = 0
        self.timestamp = time.time()
        self.size = 0

        # Register as Transaction Manager so my transaction hooks get called.
        get_transaction().register(self)


    def remove(self):
        """ Delete the cache file from disk """
        self.op_remove = 1


    def write(self, data):
        """ Write a cache file to disk """
        try:
            # Open a tempfile to safely dump data into
            if self.tmpdir:
                fd, tempname = tempfile.mkstemp(dir=self.tmpdir)
            else:
                fd, tempname = tempfile.mkstemp()
            self.temppath = tempname

            # isinstance won't work on extension class
            if type(data) == type(''):
                os.write(fd, data)
                self.size = len(data)
            else:
                # Image pdata objects?
                # there's no interface to check
                # and I don't want to test meta_type, so
                # assume that's what we have here.
                while data is not None and hasattr(data, 'next'):
                    os.write(fd, data.data)
                    self.size = self.size + len(data.data)
                    data = data.next

            os.fsync(fd)
            os.close(fd)

            self.op_write = 1

        except IOError, msg:
            zLOG.LOG( 'FileCacheManager'
                    , zLOG.ERROR
                    , 'IOError writing file'
                    , error=msg
                    )


    #######################################################
    # ZODB Transaction hooks
    #######################################################

    def commit(self, reallyme, t):
        """ Called for every (sub-)transaction commit """
        pass


    def tpc_begin(self, transaction):
        """ Called at the beginning of a transaction """
        pass


    def tpc_abort(self, transaction):
        """ Called on abort - but not always :( """
        pass


    def abort(self, reallyme, t):
        """ Called if the transaction has been aborted """
        if self.op_write and self.temppath:
            # If this was a write operation we need to remove the 
            # tempfile that's still hanging around.
            try:
                os.unlink(self.temppath)
            except (OSError, IOError):
                pass

        self.temppath = None
        self.op_write = 0
        self.op_remove = 0


    def tpc_vote(self, transaction):
        """ Only called for real transactions, not subtransactions """
        self._transaction_done = 1


    def tpc_finish(self, transaction):
        """ Called at the end of a successful transaction """
        if getattr(self, '_transaction_done', 0) and self.op_remove:
            self.cache.record(self)
            try:
                try:
                    self.writelock.acquire()
                    if os.path.exists(self.path): # XXX race?
                        os.remove(self.path)
                        if _debug:
                            print "removed %s" % self.path
                except IOError:
                    zLOG.LOG( 'FileCacheManager'
                              , zLOG.ERROR
                              , 'IOError removing file "%s"' % self.path
                              , error=sys.exc_info()
                              )
            finally:
                self.writelock.release()
                self.op_remove = 0
                
        elif getattr(self, '_transaction_done', 0) and self.op_write:
            self.cache.record(self)
            # Now rename the tempfile to reflect the desired name.
            # This may fail if they are not on the same filesystem.
            try:
                try:
                    self.writelock.acquire()
                    parent = os.path.realpath(os.path.dirname(self.path))
                    if not os.path.isdir(parent) and os.path.exists(parent):
                        zLOG.LOG(
                            'FileCacheManager',
                            zLOG.ERROR,
                            'IOError renaming file "%s" to "%s"' % (
                            self.temppath, self.path),
                            'The destination path exists'
                            )
                        return
                    
                    if self.makedirs and not os.path.isdir(parent):
                        try:
                            os.makedirs(parent)
                        except IOError:
                            zLOG.LOG(
                                'FileCacheManager',
                                zLOG.ERROR,
                                'IOError making dirs for file "%s"'% self.path,
                                error=sys.exc_info()
                                )
                            return
                            
                    # XXX Windows will fail here if target path exists, shrug
                    os.rename(self.temppath, self.path)
                    if _debug:
                        print "wrote %s" % self.path
                except IOError:
                     zLOG.LOG( 'FileCacheManager'
                             , zLOG.ERROR
                             , 'IOError renaming file "%s" to "%s"' % (
                                        self.temppath, self.path)
                             , error=sys.exc_info()
                             )
            finally:
                self.writelock.release()
                self.temppath = None
                self.op_write = 0


    def sortKey(self, *ignored):
        """ The sortKey method is used for recent ZODB compatibility which
            needs to have a known commit order for lock acquisition.  We use
            our timestamp, which will ensure that the order in which file
            cache entries are created will be the way the operations they
            imply are played back.
        """
        return self.timestamp
