##############################################################################
#
# Copyright (c) 2004 Chris McDonough, Paul Winkler, Jens Vagelpohl and
# Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os
import time
from thread import allocate_lock

from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import view_management_screens
from Globals import InitializeClass
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.StandardCacheManagers.RAMCacheManager import RAMCacheManager

from FileCache import FileCache
from permissions import change_cache_managers

caches = {}
cache_lock = allocate_lock()

class FileCacheManager(RAMCacheManager):
    """ A cache manager for caching data to filesystem """
    security = ClassSecurityInfo()
    meta_type = 'File Cache Manager'
    manage_fs = PageTemplateFile('www/editFCM', globals())

    manage_options = (
        ( { 'label' : 'File System', 'action' : 'manage_fs',
            'help' : ('FileCacheManager', 'Edit.stx') }
          ,
        ) + RAMCacheManager.manage_options
        )

    _tempfile_path = ''

    manage_stats = PageTemplateFile('www/statsFCM', globals())
    manage_main = PageTemplateFile('www/propsFCM', globals())

    def __init__(self, ob_id, path='/tmp', tpath='', title=''):
        # based on RAMCacheManager
        self.id = ob_id
        self._dir = path
        self._tempfile_path = tpath
        self.title = title
        self._naming_expr_text = ''
        self._settings = {
            'threshold': 1000,
            'cleanup_interval': 3600,
            'request_vars': ('AUTHENTICATED_USER',),
            'max_age': 3600,
            }
        self.__cacheid = '%s_%f' % (id(self), time.time())

    def getCacheReport(self):
        """
        Reports on the contents of the cache (used by Statistics tab).
        """
        cache = self.ZCacheManager_getCache()
        return cache.values()
    
    def manage_editProps(self, title, settings=None, REQUEST=None):
        'Changes the cache settings.'
        if settings is None:
            settings = REQUEST
        self.title = str(title)
        if REQUEST is not None:
            return self.manage_main(
                self, REQUEST, manage_tabs_message='Properties changed.')

    def ZCacheManager_getCache(self):
        """ Cache Settings """
        cacheid = self.__cacheid
        try:
            cache_lock.acquire()
            try:
                return caches[cacheid]
            except KeyError:
                cache = FileCache()
                cache.setDir(self._dir)
                cache.setTempfileDir(self._tempfile_path)
                cache.setNamingExpression(self._naming_expr_text)
                caches[cacheid] = cache

                return cache
        finally:
            cache_lock.release()


    security.declareProtected(view_management_screens, 'getDir')
    def getDir(self):
        """ Return the cache directory path """
        cache = self.ZCacheManager_getCache()

        return cache.getDir()

    security.declareProtected(view_management_screens, 'getTempfileDir')
    def getTempfileDir(self):
        """ Return the temporary directory path """
        cache = self.ZCacheManager_getCache()

        return cache.getTempfileDir()

    security.declareProtected(change_cache_managers, 'setDir')
    def setDir(self, path):
        """ Set the cache directory path """
        cache = self.ZCacheManager_getCache()
        cache.setDir(path)
        self._dir = path

    security.declareProtected(change_cache_managers, 'setDir')
    def setTempfileDir(self, path):
        """ Set the tempfile directory path """
        cache = self.ZCacheManager_getCache()
        cache.setTempfileDir(path)
        self._tempfile_path = path

    security.declareProtected(view_management_screens, 'getNamingExpression')
    def getNamingExpression(self):
        """ Retrieve the TAL expression for naming file system files """
        return self._naming_expr_text

    security.declareProtected(change_cache_managers, 'setNamingExpression')
    def setNamingExpression(self, naming_expression_text=''):
        """ Set a naming expression for filesystem files """
        cache = self.ZCacheManager_getCache()
        cache.setNamingExpression(naming_expression_text)
        self._naming_expr_text = naming_expression_text

    security.declareProtected(change_cache_managers, 'manage_edit')
    def manage_edit(self, path='/tmp', tpath='',
                    naming_expression_text='', REQUEST=None):
        """ Edit the filesystem-related values """
        self.setNamingExpression(naming_expression_text)
        self.setDir(path)
        self.setTempfileDir(tpath)

        if REQUEST is not None:
            return self.manage_fs(
                self, REQUEST, manage_tabs_message='Changes saved')


InitializeClass(FileCacheManager)


manage_addFileCacheManagerForm = PageTemplateFile('www/addFCM', globals())

def manage_addFileCacheManager(self, id, path='/tmp', tpath='',
                               title='', REQUEST=None):
    """ Adds a FileCacheManager to the folder. """
    # Test to see that we can write into the place we got as the cache dir
    self._setObject(id, FileCacheManager(id, path, tpath, title))

    if REQUEST is not None:
        REQUEST.set('manage_tabs_message', 'File Cache Manager created')
        return self.manage_main(self, REQUEST)

