##############################################################################
#
# Copyright (c) 2004 Chris McDonough, Paul Winkler, Jens Vagelpohl and
# Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os
import threading
import time

import ZODB
from ZODB.MappingStorage import MappingStorage
from ZEO.ClientStorage import ClientStorage
from ZPublisher.Iterators import filestream_iterator
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import noSecurityManager
from AccessControl.User import UnrestrictedUser
from Globals import package_home
from Testing import makerequest
from OFS.Application import Application
from OFS.Folder import Folder, manage_addFolder
from OFS.Image import manage_addImage
from Products.PythonScripts.PythonScript import manage_addPythonScript

from Products.FileCacheManager import FileCacheManager

FCM_ID = 'fcm1'
PYSCRIPT_ID = 'ps1'
ROOT = None
IMAGES = {}
PROD_DIR = os.path.join(os.getcwd(), 'Products', 'FileCacheManager')
IMAGE_DIR = os.path.join(PROD_DIR, 'tests', 'images')

for test_file in os.listdir(IMAGE_DIR):
    full_path = os.path.join(IMAGE_DIR, test_file)
    if os.path.isfile(full_path):
        fh = open(full_path, 'rb')
        IMAGES[test_file] = (fh.read() * 100)
        fh.close()

######################
# Tweakable items
FCM_DIR = '/tmp/foo'
FCM_NAMING_EXPRESSION = ''
PYSCRIPT_PARAMS = ''
PYSCRIPT_BODY = """
"""
READERS=1
WRITERS=0
ITERATIONS = 10
# End tweakable items
######################


def setUp():
    # First, create a ZODB for the ZODB-based objects
    #s = MappingStorage()
    s = ClientStorage(('127.0.0.1', 18001))
    db = ZODB.DB(s)
    jar = db.open()
    app_root = jar.root()
    app_root['app'] = Application()
    global ROOT
    ROOT = app_root.get('app')

    # Set up a Security Manager so we do everything as a manager
    newSecurityManager(None, UnrestrictedUser('manager', '', ['Manager'], []))

    # Now stick a FileCacheManager object into our testing root
    factory = FileCacheManager.manage_addFileCacheManager
    factory(ROOT, FCM_ID, path=FCM_DIR, title='Testing FCM')
    fcm = getattr(ROOT, FCM_ID)

    # Set up a naming expression from FCM_NAMING_EXPRESSION above
    fcm.setNamingExpression(FCM_NAMING_EXPRESSION)
    assert fcm.getNamingExpression() == FCM_NAMING_EXPRESSION

    # Set up a Script(Python) to be used in the naming expression
    manage_addPythonScript(ROOT, PYSCRIPT_ID)
    pyscript = getattr(ROOT, PYSCRIPT_ID)
    pyscript.ZPythonScript_edit(PYSCRIPT_PARAMS, PYSCRIPT_BODY)
    assert pyscript.params() == PYSCRIPT_PARAMS
    assert pyscript.body() == PYSCRIPT_BODY.strip()

    # Set up a series of folders for storing content
    # and store content in it.
    for i in range(1, 11):
        folder_id = 'folder_%d' % i
        manage_addFolder(ROOT, folder_id)
        folder_ob = getattr(ROOT, folder_id)

        for id, content in IMAGES.items():
            manage_addImage(folder_ob, id, content)

        assert len(folder_ob.objectIds()) == len(IMAGES.keys())

    assert len(ROOT.objectIds(['Folder'])) == 10

    # Now find and associate cacheable objects
    # That ZCacheable API is a PITA
    cacheables = fcm.ZCacheManager_locate(0, 1, meta_types=['Image', 'File'])
    to_be_cached = {}
    for cacheable in cacheables:
        to_be_cached['associate_%s' % cacheable['path']] = 1
    fcm.ZCacheManager_setAssociations(props=to_be_cached)

    # Remove the Security Manager for now
    noSecurityManager()


def tearDown():
    for root, dirs, files in os.walk(FCM_DIR, topdown=False):
        for name in files:
            os.unlink(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name)) 


class ReadStresser(threading.Thread):
    """ Objects that simulate read or write access to the cached resources """
    def __init__(self, number, root, paths, iterations):
        self.number = number
        self.root = root
        self.paths = paths
        self.iterations = iterations
        self.finished = 0
        threading.Thread.__init__(self)

    def run(self):
        counter = 0
        n_counter = 0
        e_counter = 0
        c_counter = 0
        for i in range(self.iterations):
            counter += 1
            for path in self.paths:
                root = makerequest.makerequest(self.root)
                ob = root.unrestrictedTraverse(path)
                #print '.',
                sys.stdout.flush()
                data = ob.index_html(root.REQUEST, root.REQUEST['RESPONSE'])
                original_data = IMAGES.get(ob.getId())

                if isinstance(data, filestream_iterator):
                    data_str = ''
                    while 1:
                        try:
                            data_str = data_str + data.next()
                        except StopIteration:
                            break
                    data = data_str

                if not data:
                    print 'N',
                    sys.stdout.flush()
                    n_counter += 1
                if data != original_data:
                    print 'E',
                    sys.stdout.flush()
                    e_counter += 1
                if ob.ZCacheable_get(default=None) is None:
                    print 'C',
                    sys.stdout.flush()
                    c_counter += 1

        print '\n\nReader %d: %d total / %d empty / %d faulty / %d empty cache' % ( self.number
                                                                                  , counter
                                                                                  , n_counter
                                                                                  , e_counter
                                                                                  , c_counter
                                                                                  )
        sys.stdout.flush()
        self.finished = 1

    def isFinished(self):
        return self.finished


class WriteStresser(threading.Thread):
    """ Objects that simulate read or write access to the cached resources """
    def __init__(self, number, root, paths, iterations):
        self.number = number
        self.root = root
        self.paths = paths
        self.iterations = iterations
        self.finished = 0
        threading.Thread.__init__(self)

    def run(self):
        for i in range(self.iterations):
            for path in self.paths:
                root = makerequest.makerequest(self.root)
                ob = root.unrestrictedTraverse(path)
                #print ':',
                sys.stdout.flush()
                original_data = IMAGES.get(ob.getId())
                ob.manage_upload(file=original_data)

        self.finished = 1

    def isFinished(self):
        return self.finished

def numActive(threads):
    count = 0
    for thread in threads:
        if not thread.isFinished():
            count += 1

    return count

def stress():
    readers = []
    writers = []
    paths = []

    for id in ROOT.objectIds(['Folder']):
        for img_file in IMAGES.keys():
            paths.append('%s/%s' % (id, img_file))

###
#    for i in range(ITERATIONS):
#        for path in paths:
#            root = makerequest.makerequest(ROOT)
#            ob = root.unrestrictedTraverse(path)
#            print '.',
#            sys.stdout.flush()
#            data = ob.index_html(root.REQUEST, root.REQUEST['RESPONSE'])
#
#            if isinstance(data, filestream_iterator):
#                data_str = ''
#                while 1:
#                    try:
#                        data_str = data_str + data.next()
#                    except StopIteration:
#                        break
#                        
#                data = data_str
#
#            if not data:
#                print 'N',
#                self.stdout.flush()
#            if data != IMAGES.get(ob.getId()):
#                print 'E',
#                sys.stdout.flush()
#            if ob.ZCacheable_get(default=None) is None:
#                print 'C',
#                sys.stdout.flush()
###

    for i in range(READERS):
        thread = ReadStresser(i, ROOT, paths, ITERATIONS)
        readers.append(thread)

    for i in range(WRITERS):
        thread = WriteStresser(i, ROOT, paths, ITERATIONS)
        writers.append(thread)
        
    for reader in readers:
        reader.start()
        time.sleep(0.1)

    for writer in writers:
        writer.start()
        time.sleep(0.1)

    while numActive(readers) + numActive(writers) > 0:
        time.sleep(1)
        
def run():
    #print 'Cleanup...'
    #sys.stdout.flush()
    #tearDown()

    print 'Setting up...'
    sys.stdout.flush()
    setUp()
    
    print 'Stressing...'
    sys.stdout.flush()
    stress()

    #print 'Tearing down...'
    #sys.stdout.flush()
    #tearDown()


if __name__ == '__main__':
    run()

