##############################################################################
#
# Copyright (c) 2004 Chris McDonough, Paul Winkler, Jens Vagelpohl and
# Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os, unittest, warnings
import random
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from ZPublisher.Iterators import filestream_iterator
from Products.PageTemplates.PythonExpr import PythonExpr
from Products.FileCacheManager.FileCacheManager import FileCache, \
     FileCacheManager

FCM_ID = 'fcm1'
FCM_TITLE = 'Testing FileCacheManager'
FCM_DIR = '/tmp/foo'


class Dummy:
    """ fake content object """
    security = ClassSecurityInfo()

    def __init__(self, id, fakepath, size):
        self.id = id
        randchar = chr(random.randint(0, 255))
        self.data = randchar * size
        self.aq_explicit = self
        self.fakepath = fakepath

    security.declarePublic('getPhysicalPath')
    def getPhysicalPath(self):
        """ """
        fakepath = self.fakepath.split('/')
        fakepath.append(self.id)
        return fakepath

    security.declarePublic('getId')
    def getId(self):
        """ """
        return self.id

    security.declarePublic('absolute_url')
    def absolute_url(self, relative=0):
        """ """
        return self.fakepath

InitializeClass(Dummy)
    

class FileCacheTestBase(unittest.TestCase):
    def setUp(self):
        self.FC = FileCache(path=FCM_DIR)
        self.f1 = Dummy(id='f1', fakepath='/spam/eggs', size=30)
        self.f2 = Dummy(id='f2', fakepath='/bacon', size=1000 * 1000)
        self.all_files = (self.f1, self.f2)
        get_transaction().commit()
        
    def tearDown(self):
        for root, dirs, files in os.walk(FCM_DIR, topdown=False):
            for name in files:
                os.unlink(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))


class FileCacheTests(FileCacheTestBase):

    def testInstantiation(self):
        self.assertEqual(self.FC.getDir(), FCM_DIR)
        self.assertEqual(self.FC._naming_expr, None)

    def testGetSetDir(self):
        self.assertEqual(self.FC.getDir(), FCM_DIR)
        self.FC.setDir('/tmp')
        self.assertEqual(self.FC.getDir(), '/tmp')

    def testGetSetTempfileDir(self):
        self.assertEqual(self.FC.getTempfileDir(), '')
        self.FC.setTempfileDir('/tmp')
        self.assertEqual(self.FC.getTempfileDir(), '/tmp')

    def testGetSetNamingExpression(self):
        # By default, no naming expression is used
        self.assertEqual(self.FC._naming_expr, None)

        # Set a valid expression
        expr = "python: '%s.txt' % (object.getId())"
        self.FC.setNamingExpression(expr)
        self.assert_(self.FC._naming_expr != None)
        self.assert_(isinstance(self.FC._naming_expr, PythonExpr))

    def testZCache_set(self):
        # First, test against the standard non-TAL-Expression munger
        for ob in self.all_files:
            self.FC.ZCache_set(ob)
            get_transaction().commit()
            # don't know how to verify that it worked
            # without testing the implementation.
            fpath = self.FC._fileName(ob)
            self.failUnless(os.path.exists(fpath))

        # Now, add a TAL naming expression
        self.FC.setNamingExpression("python: object_url.replace('/', '_')")
        for ob in self.all_files:
            self.FC.ZCache_set(ob)
            get_transaction().commit()
            fpath = self.FC._fileName(ob)
            self.failUnless(os.path.exists(fpath))

    def testZCache_set_aborted(self):
        # First, test against the standard non-TAL-Expression munger
        for ob in self.all_files:
            self.FC.ZCache_set(ob)
            get_transaction().abort()
            # don't know how to verify that it worked
            # without testing the implementation.
            fpath = self.FC._fileName(ob)
            self.failIf(os.path.exists(fpath))
            
    def testZCache_get(self):
        """ assume that ZCache_set works :-P """

        for ob in self.all_files:
            self.FC.ZCache_set(ob) # assume that works.
            get_transaction().commit()

            # we should get back a filestream_iterator.
            file_iter = self.FC.ZCache_get(ob)
            self.failUnless(isinstance(file_iter, filestream_iterator))

            # getting all data from the filestream_iterator ...
            all_data = []
            while 1:
                try:
                    all_data.append(file_iter.next())
                except StopIteration:
                    break

            all_data = ''.join(all_data)
            # it should be the right data.
            self.assertEqual(ob.data, all_data)
            
    def testZCache_invalidate(self):
        for ob in self.all_files:
            self.FC.ZCache_set(ob)
            get_transaction().commit()
            fpath = self.FC._fileName(ob)            
            self.FC.ZCache_invalidate(ob)
            get_transaction().commit()
            # the file should now be gone
            self.failIf(os.path.exists(fpath))

    def testZCache_invalidate_aborted(self):
        for ob in self.all_files:
            self.FC.ZCache_set(ob)
            get_transaction().commit()
            fpath = self.FC._fileName(ob)            
            self.FC.ZCache_invalidate(ob)
            get_transaction().abort()
            # the file should *still be there*
            self.failUnless(os.path.exists(fpath))

    def test_values(self):
        self.assertEqual(self.FC.entries, {})
        for ob in self.all_files:
            self.FC.ZCache_set(ob)
            get_transaction().commit()
        self.assertNotEqual(self.FC.entries, {})
        for ob in self.all_files:
            self.FC.ZCache_invalidate(ob)
            get_transaction().commit()
        self.assertEqual(self.FC.entries, {})

class FileCacheManagerTestBase(unittest.TestCase):

    def setUp(self):
        self.FCM = FileCacheManager(FCM_ID, path=FCM_DIR, title=FCM_TITLE)

    def tearDown(self):
        for root, dirs, files in os.walk(FCM_DIR, topdown=False):
            for name in files:
                os.unlink(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))


class FileCacheManagerTests(FileCacheManagerTestBase):

    def testInstantiation(self):
        self.failUnless(self.FCM.getId() == FCM_ID)
        self.failUnless(self.FCM.getDir() == FCM_DIR)
        self.failUnless(self.FCM.title == FCM_TITLE)
        self.failUnless(self.FCM.getNamingExpression() == '')

    def testGetSetDir(self):
        self.failUnless(self.FCM.getDir() == FCM_DIR)
        self.FCM.setDir('/tmp')
        self.failUnless(self.FCM.getDir() == '/tmp')

    def testGetSetNamingExpression(self):
        # By default, no naming expression is used
        self.assertEqual(self.FCM.getNamingExpression(), '')

        # Set an invalid TAL naming expression, no change should occur
        self.assertRaises(Exception, "self.FCM.setNamingExpression('foo: 1+1')")
        self.assertEqual(self.FCM.getNamingExpression(), '')

        # Set a valid expression
        expr = "python: '%s.txt' % (object.getId())"
        self.FCM.setNamingExpression(expr)
        self.assertEquals(self.FCM.getNamingExpression(), expr)

    def testGetCache(self):
        c = self.FCM.ZCacheManager_getCache()
        self.failUnless(isinstance(c, FileCache))


def test_suite():
    suite = unittest.TestSuite((
        unittest.makeSuite(FileCacheTests),
        unittest.makeSuite(FileCacheManagerTests)
        ))
    
    return suite

def run():
    unittest.TextTestRunner().run(test_suite())

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')


